clear

% Generate a radio signal at a given noise level, with a random data set,
% and proceed to map the BER at each threshold level. The median of the
% threshold values with the lowest error rate is assumed ideal.

%%
% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency.

% Enter various constants for the simulation
sample_rate = 4;                % Multiple of the chip rate, depends on correlator design
correlator_type = 0;            % 0 = standard, 1 = voted, 2 = dual voted, 3 = variable ADC
ADC_bits = 3;                   % Only used for correlator_type = 3
desired_snr = 1.5;              % In decibels, needs to be high enough to cause some errors
data_vector_length = 1000;

%% Generate sequences
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];
barker_11_1bit = [1 1 1 0 0 0 1 0 0 1 0];

% Determine the threshold levels
switch correlator_type
   case 0
       %% Standard correlator
       % The number of discrete resolution points
       threshold_top = sample_rate * length(barker_11);
       threshold_levels_mid = floor((sample_rate * length(barker_11))/2);
       
   case 1
      %% Voted type
        if mod(sample_rate,2) == 0
           error('Sampling rate is even! Use an odd sampling rate for the voted unit.')
        end
        
        % The number of discrete resolution points
        threshold_top = length(barker_11);
        threshold_levels_mid = floor((threshold_top + 1) / 2);

   case 2
      %% Dual voted type
        % Check if the sampling rates align
        sample_rate_mini = 3;           % The sampling rate of the sub correlator, for mode 2
        if rem(sample_rate, sample_rate_mini) 
            error('Sampling rate must be multiple of sub correlator rate!')
        end
        
        % The number of discrete resolution points
        threshold_top = 2*length(barker_11);
        threshold_levels_mid = floor((threshold_top + 1) / 2);

    case 3
        %% Variable bit quantized
        error('Error! Not supported!')

    otherwise
    error('Error! Undefined correlator used!')

end

% Vector to hold each result
error_counter_vector = zeros(1, threshold_levels_mid);
data_counter_vector = zeros(1, threshold_levels_mid);

% Generate a random vector of 1 and -1, representing data to be used at
% this noise level
data = 2*randi(2, 1, data_vector_length) - 3;

% Create the reference code used in the correlator
% Expanded to match the sampling rate
pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, length(barker_11)*sample_rate);
pn_code_1bit = reshape((barker_11_1bit' * ones(1, sample_rate))', 1, length(barker_11_1bit)*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * data_vector_length;
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Create noise and scale to match desired SNR
noise = randn(1, length(tx_data));
noise = noise / norm(noise) * norm(tx_data) / 10.0^(0.05*desired_snr);

% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data_noise = tx_data + noise;
radio_data = radio_data_noise * (1- var(abs(radio_data_noise)));

% Open a worker pool for parallel computation
matlabpool('open');

poolSize = matlabpool('size');
if poolSize == 0
   error('parallel:demo:poolClosed', ...
        'This demo needs an open MATLAB pool to run.');
end
fprintf('This script is running on %d MATLABPOOL workers.\n', ...
    matlabpool('size'));
% Turn off temporary variable warnings, temporarily
warning('off', 'MATLAB:mir_warning_maybe_uninitialized_temporary');

%time the calculation
tic;

% Run the algorithm a number of times, and average the result
parfor q = 1:threshold_levels_mid

    %% Cross correlate
    % Select the correlator and thresholds to use

    switch correlator_type
       case 0
           %% Standard correlator
           
            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = standard_correlation_1bit(rx_data, sample_rate);

       case 1
          %% Voted type
            if mod(sample_rate,2) == 0
               error('Sampling rate is even! Use an odd sampling rate for the voted unit.')
            end

            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = voted_correlator_1bit(rx_data, sample_rate);

       case 2
          %% Dual voted type
            % Check if the sampling rates align
            sample_rate_mini = 3;           % The sampling rate of the sub correlator, for mode 2
            if rem(sample_rate, sample_rate_mini) 
                error('Sampling rate must be multiple of sub correlator rate!')
            else
                mini_corr_units = sample_rate / sample_rate_mini;   % How many mini corrs
            end

            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = dual_voted_correlation_1bit(rx_data, sample_rate);

        case 3
            %% Variable bit quantized
            rx_data = variable_bit_quantize( radio_data, ADC_bits );
            
            % Cross correlation of radio data and PN code
            correlation_vector = xcorr(rx_data, pn_code);
            L = length(rx_data) + length(pn_code) - 1;
            start_corr = length(correlation_vector) - L;
            correlation_vector = correlation_vector(start_corr:end);
            
        otherwise
        error('Error! Undefined correlator used!')

    end

    %% Detect data bits in the correlator vector
    detected_data = detect_correlator_peaks(correlation_vector, (threshold_top - q), q, sample_rate);

    % Count errors
    error_counter_vector(q) = 0;
    if (length(detected_data) > length(data))
        for l = 1:length(data)
            if (detected_data(l) ~= data(l))
                error_counter_vector(q) = error_counter_vector(q) + 1;
            end
        end
    else
        for l = 1:length(detected_data)
            if (detected_data(l) ~= data(l))
                error_counter_vector(q) = error_counter_vector(q) + 1;
            end
        end
    end
    
    % Keep track of how many data bits are actual found
    data_counter_vector(q) = length(detected_data);
    
    % Keep the user awake
    disp(['Threshold lower ' num2str(q) ' and upper ' num2str((threshold_top - q)) ' of ' num2str(threshold_top) ' complete.'])
    disp(['Errors for lower thres: ' num2str(q) ' are ' num2str(error_counter_vector(q))])
end

TimeSpent = toc;

% Find the sweet spot, feasible values will be for thresholds where the the
% number of detected bits is close to the actual sent bit count
r = 1;
expected_data_len_max = 1.1 * data_vector_length;
expected_data_len_min = 0.9 * data_vector_length;
for j = 1:threshold_levels_mid
    if ((data_counter_vector(j) >= expected_data_len_min) && (data_counter_vector(j) < expected_data_len_max))
        pruned_error_vector(r) = error_counter_vector(j);
        pruned_datalen_vector(r) = data_counter_vector(j);
        threshold_index(r) = j;
        r = r + 1;
    end
end

% Normalize the vectors
pruned_error_vector = pruned_error_vector/norm(pruned_error_vector);
pruned_datalen_vector = pruned_datalen_vector/norm(pruned_datalen_vector);

%% Display results
switch correlator_type
    case 0
        str_corr = 'Standard correlator';
	case 1
        str_corr = 'Voted correlator';
    case 2
        str_corr = 'Dual voted correlator';
    case 3
        str_corr = ['Variable quantized correlator, ADC bits: ' num2str(ADC_bits)];
    otherwise
        str_corr = 'Undefined correlator.';
end
disp(str_corr)

disp(['Sample rate: ' num2str(sample_rate)])
disp(['Error vector: ' num2str(error_counter_vector)])
disp(['Expected length is: ' num2str(data_vector_length) ' Detected data length vector ' num2str(data_counter_vector)])
disp(['Pruned Error vector: ' num2str(pruned_error_vector)])
disp(['Pruned Data length vector: ' num2str(pruned_datalen_vector)])
disp(['Threshold indices in pruned vectors: ' num2str(threshold_index)])
disp(['Calculation time: ' num2str(TimeSpent) ' seconds.'])

figure(1)
hold on;
plot(threshold_index, pruned_error_vector, 'r');
plot(threshold_index, pruned_datalen_vector);
legend('Errors', 'Data detected');
hold off;
title([str_corr '. Normalized error count and data length, for sampling rate: ' num2str(sample_rate) ' SNR: ' num2str(desired_snr)])
xlabel('Lower threshold value')

matlabpool('close');
% Turn on warnings again
warning('on', 'MATLAB:mir_warning_maybe_uninitialized_temporary');
