% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency, and correlated
% with a local PN code (Barker-11). The correlator is a matched filter
% (parallel correlator), using several smaller, simpler, correlators. The
% output of these are added together to form the final correlation measure.

% Enter various constants for the simulation
noise_factor = 0.9;             % Set the amplitude of the noise
sample_rate = 6;                % Multiple of the chip rate, NB! must be multiple of two
noise_resolution = 100;         % How many samples of noise for every sample of data
data = [1 1 -1 -1 -1 1 1 -1 -1 -1 -1 1 1 1];   % Data sequence to transmitt
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];     % Needs sharp autocorrelation curve
barker_11_1bit = [1 1 1 0 0 0 1 0 0 1 0];

% Parameters for the multi-voter
sample_rate_mini = 2;           % Stuck at 2, due to clock constraints
mini_corr_units = sample_rate / sample_rate_mini;   % How many mini corrs

% Generate sequences
% Create the reference code used in the correlator
% Expanded to match the sampling rate
pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, length(barker_11)*sample_rate);
pn_code_1bit = reshape((barker_11_1bit' * ones(1, sample_rate))', 1, length(barker_11_1bit)*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * length(data);
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Mix noise into the data stream, and read again using the ADC
% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);
% Create a vector of noise
noise = noise_factor*randn(1, length(tx_data)*noise_resolution);
% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data = tx_temp + noise;
radio_snr = 10 * log10(mean(tx_temp.^2) / mean(noise.^2));
radio_data = radio_data * (1- var(abs(radio_data)));

% rx_data is radio_data sampled at every x sample.
rx_data = radio_data(1:noise_resolution:length(radio_data));
temp = rx_data;
% Next the data is quantized to allowable levels
error_counter = 0;
for i = 1 : length(rx_data)
    % Quantize each sample
    if rx_data(i) >= 0
        rx_data(i) = 1;
    else
        rx_data(i) = 0;
    end
    
    % Count how many errors are present
    if (rx_data(i) ~= tx_data(i))
        if ~((tx_data(i) == -1) && (rx_data(i) == 0))
             error_counter = error_counter + 1;
        end
    end
end

% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
legend('Radio signal', 'TX Data');
hold off;
title(['Received analog data from radio link with AGC. SNR: ' num2str(radio_snr) ' dB'])

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(temp, 'g');
stem(rx_data, 'r');
legend('TX Data', 'Analog Sample', 'Digital Sample');
hold off;
title(['Sampled 1 bit data from radio link, error percent: ' num2str((error_counter / length(rx_data))*100) '%'])

% Sliding correlation of the received signal with the PN code. Should
% get nice positive and negative peaks for each bit.
figure(2)

% Expected data to receive
subplot(3, 1, 1);
%data_for_stem = zeros(1, (2 + length(data)));
%data_for_stem(2:1+length(data)) = data;
%data_for_stem(1) = 0;
data_for_stem = zeros(1, (1 + length(data)));
data_for_stem(1:length(data)) = data;
stem(data_for_stem);
title('Expected data sequence')

% Cross correlate
% Matched filter algorithm
% Create the ADC register, fill with noise
input_reg = randi(2,1,length(pn_code_1bit));
input_reg = input_reg - 1;
% Flip the PN code from right to left, so we xcorr and not convolute
pn_code_1bit = fliplr(pn_code_1bit);
output_standard = zeros(1, length(rx_data));
for k = 1:length(rx_data)
    % Shift in new data
    input_reg(2:end) = input_reg(1:end-1);
    input_reg(1) = rx_data(k);
    
    % The current standard approach
    for l = 1:length(pn_code_1bit)
        if (pn_code_1bit(l) == input_reg(l))
            output_standard(k) = output_standard(k) + 1;
        end
    end
end

% Stuff for the multi corr
pn_code_1bit_mini = reshape((barker_11_1bit' * ones(1, sample_rate_mini))', 1, length(barker_11_1bit)*sample_rate_mini);
pn_code_1bit_mini = fliplr(pn_code_1bit_mini);

% Downsample the received data into two registers, each holding unique data
sub_corr_length = length(rx_data)/mini_corr_units;
downsampled_rxdata = zeros(mini_corr_units, sub_corr_length);
for j = 1:mini_corr_units
    downsampled_rxdata(j, 1:end) = rx_data(j : mini_corr_units : end);
end
% Sub correlation registers
sub_corr = zeros(mini_corr_units, sub_corr_length);
% Input registers for each sub correlator
input_reg_mini = randi(2,mini_corr_units,length(pn_code_1bit_mini));
input_reg_mini = input_reg_mini - 1;

% Run through every input sample
for k = 1:sub_corr_length
    
    % For each sub correlator
    for j = 1:mini_corr_units
        % Shift in new data
        input_reg_mini(j, 2:end) = input_reg_mini(j, 1:end-1);
        input_reg_mini(j, 1) = downsampled_rxdata(j, k);
        % Regular correlation for each sub correlator
        for l = 1:length(pn_code_1bit_mini)
            if (pn_code_1bit_mini(l) == input_reg_mini(j, l))
                sub_corr(j, k) = sub_corr(j, k) + 1;
            end
        end
    end
end

% Up sample each mini-corr...
mini_corr = zeros(mini_corr_units, sub_corr_length*mini_corr_units);
for j = 1:mini_corr_units
	mini_corr(j, :) = reshape((sub_corr(j, 1:end)' * ones(1, mini_corr_units))', 1, sub_corr_length*mini_corr_units);
end
% ...shift them apart...
for j = 2:mini_corr_units
    mini_corr(j, j:end) = mini_corr(j, 1:end-(j-1));
    mini_corr(j, 1:(j-1)) = 0;
end
% ...and add together
output_test_multi = zeros(1, sub_corr_length*mini_corr_units);
for j = 1:mini_corr_units
	output_test_multi = output_test_multi + mini_corr(j, :);
end


% Create threshold lines
upper_thres = 5 * sample_rate * ones(1, length(rx_data));
lower_thres = -5 * sample_rate * ones(1, length(rx_data));
upper_thres = (upper_thres + sample_rate * 11)/2;
lower_thres = (lower_thres + sample_rate * 11)/2;

subplot(3, 1, 2);
hold on;
plot(output_standard);
plot(upper_thres, 'r');
plot(lower_thres, 'r');
hold off;
title(['Current x-corr standard, sample rate ' num2str(sample_rate) ' times chip frequency.'])
xlabel('Sample number')
 
subplot(3, 1, 3);
hold on;
plot(output_test_multi);
plot(upper_thres, 'r');
plot(lower_thres, 'r');
hold off;
title(['With ' num2str(mini_corr_units) ' "two times chip rate" correlators summed.'])
xlabel('Sample number')

% Write the received data vector to a file, for use in a testbench
str1 = ['Data sequence: [' num2str(data) ']'];
str2 = ['Signal SNR: ' num2str(radio_snr) 'dB. '];
str3 = ['Sampling rate: ' num2str(sample_rate)];
str4 = 'Received data vector from 1-bit ADC:';

fName = 'rx_vect_for_tb.txt';
fid = fopen(fName,'w');
if fid ~= -1
  fprintf(fid,'%s\r\n',str1);
  fprintf(fid,'%s\r\n',str2); 
  fprintf(fid,'%s\r\n',str3);
  fprintf(fid,'%s\r\n',str4);
  fclose(fid);                   
end
dlmwrite(fName, rx_data', '-append',... 
         'delimiter','\t',...
         'newline','pc');
