% Generate a radio signal with a random data set, while increasing the noise
% level. Noise if first increased in large steps, until an error is
% detected. Then the algorithm jumps back two steps, and repeats with a
% smaller step size. The result once an error is detected now is stored.
% The process is repeated a number of times, and the final average is
% taken.

%%
% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency.

% Enter various constants for the simulation
sample_rate = 4;                % Multiple of the chip rate, depends on correlator design
correlator_type = 0;            % 0 = standard, 1 = voted, 2 = dual voted, 3 = variable ADC
ADC_bits = 3;                   % Only used for correlator_type = 3
noise_increment_fine = 0.005;         % Set the fine simulation resolution
noise_increment_rough = 0.1;         % Set the rough simulation resolution
averaging_passes = 20;
data_vector_length = 1000;

%% Generate sequences
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];
barker_11_1bit = [1 1 1 0 0 0 1 0 0 1 0];

% Vector to hold each result
SNR_vector = zeros(1, averaging_passes);
error_counter_vector = zeros(1, averaging_passes);
noise_factor_vector = zeros(1, averaging_passes);

% Open a worker pool for parallel computation
matlabpool('open');

poolSize = matlabpool('size');
if poolSize == 0
   error('parallel:demo:poolClosed', ...
        'This demo needs an open MATLAB pool to run.');
end
fprintf('This demo is running on %d MATLABPOOL workers.\n', ...
    matlabpool('size'));
% Turn off temporary variable warnings, temporarily
warning('off', 'MATLAB:mir_warning_maybe_uninitialized_temporary');

%time the calculation
tic;

% Run the algorithm a number of times, and average the result
parfor q = 1:averaging_passes
    % Starting point for the noise level
    noise_factor = 0.0;             % Set the start amplitude of the noise
    noise_end = 1.5;                % Stopping point for the simulation
    % Start the simulation with rough increments, to save time
    noise_increment = noise_increment_rough;
    errors = 0;
    
    while((errors == 0) && (noise_factor < noise_end))
    % Generate a random vector of 1 and -1, representing data to be used at
    % this noise level
    data = 2*randi(2, 1, data_vector_length) - 3;

    % Create the reference code used in the correlator
    % Expanded to match the sampling rate
    pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, length(barker_11)*sample_rate);
    pn_code_1bit = reshape((barker_11_1bit' * ones(1, sample_rate))', 1, length(barker_11_1bit)*sample_rate);

    % Create the data sequence
    % Expand the data to be sent with the normal or inverted PN code
    tx_length = length(pn_code) * data_vector_length;
    tx_data = reshape((data' * pn_code)', 1, tx_length);

    % Mix noise into the data stream, and read again using the ADC
    % Create a vector of noise
    % Increase the noise with each pass
    noise_factor = noise_factor + noise_increment;
    noise = noise_factor*randn(1, length(tx_data));
    % Noise is assumed to be added to the signal. The receiver will likely have
    % an Automatic-Gain-Control stage, normalizing the signal amplitude
    radio_data = tx_data + noise;
    radio_snr = 10 * log10(mean(tx_data.^2) / mean(noise.^2));
    radio_data = radio_data * (1- var(abs(radio_data)));

    %% Cross correlate
    % Select the correlator and thresholds to use

    switch correlator_type
       case 0
           %% Standard correlator
           
            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = standard_correlation_1bit(rx_data, sample_rate);

            % Thresholds have been tuned
            lower_thres = floor(3 * sample_rate);
            upper_thres = (sample_rate * 11) - lower_thres;
            
       case 1
          %% Voted type
            if mod(sample_rate,2) == 0
               error('Sampling rate is even! Use an odd sampling rate for the voted unit.')
            end

            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = voted_correlator_1bit(rx_data, sample_rate);

            % Thresholds have been tuned
            upper_thres = 9;
            lower_thres = 2;

       case 2
          %% Dual voted type
            % Check if the sampling rates align
            sample_rate_mini = 3;           % The sampling rate of the sub correlator, for mode 2
            if rem(sample_rate, sample_rate_mini) 
                error('Sampling rate must be multiple of sub correlator rate!')
            else
                mini_corr_units = sample_rate / sample_rate_mini;   % How many mini corrs
            end

            % Quantized to 1-bit
            rx_data = one_bit_quantize(radio_data);
            % Correlate
            correlation_vector = dual_voted_correlation_1bit(rx_data, sample_rate);

            % Thresholds have been tuned
            upper_thres = 18;
            lower_thres = 4;

        case 3
            %% Variable bit quantized
            rx_data = variable_bit_quantize( radio_data, ADC_bits );
            
            % Cross correlation of radio data and PN code
            correlation_vector = xcorr(rx_data, pn_code);
            L = length(rx_data) + length(pn_code) - 1;
            start_corr = length(correlation_vector) - L;
            correlation_vector = correlation_vector(start_corr:end);

            % Determine thresholds
            thres_level = 2.8;
            upper_thres = thres_level * sample_rate * ones(1, length(correlation_vector));
            lower_thres = -thres_level * sample_rate * ones(1, length(correlation_vector));
            
        otherwise
        error('Error! Undefined correlator used!')

    end
    
    % Count sign errors between the received data and the sent data
    % Rough measure of the nosie impact
    error_counter = 0;
    for i = 1 : length(rx_data)
        % Count how many errors are present
        if sign(rx_data(i)) ~= sign(tx_data(i))
            error_counter = error_counter + 1;
        end
    end

    %% Detect data bits in the correlator vector
    detected_data = detect_correlator_peaks(correlation_vector, upper_thres, lower_thres, sample_rate);

    % Count errors
    errors = 0;
    for l = 1:length(detected_data)
        if (detected_data(l) ~= data(l))
            errors = errors + 1;
        end
    end
    
    % Check if this was the first pass, if so use finer resolution and try
    % again
    if ((errors ~= 0) && (noise_increment == noise_increment_rough))
        noise_factor = noise_factor - 2*noise_increment_rough;
        errors = 0;
        noise_increment = noise_increment_fine;
    end

    end
    
    % Store results from the pass
    SNR_vector(q) = radio_snr;
    error_counter_vector(q) = error_counter;
    noise_factor_vector(q) = noise_factor;
    
    % Keep the user awake
    disp(['Pass ' num2str(q) '. SNR: ' num2str(radio_snr) 'dB, Errors: ' num2str(error_counter) ' and noise factor: ' num2str(noise_factor)])
end

TimeSpent = toc;

% Calculate the averages
final_radio_snr = mean(SNR_vector);
final_error_counter = mean(error_counter_vector);
final_noise_factor = mean(noise_factor_vector);

% Calculate standard deviation of each parameter
snr_std_deviation = std(SNR_vector);
error_counter_std_deviation = std(error_counter_vector);
noise_factor_std_deviation = std(noise_factor_vector);

%% Display results
error_percent = 100 * (final_error_counter / (length(barker_11) * sample_rate * data_vector_length));
switch correlator_type
    case 0
        disp('Standard correlator')
	case 1
        disp('Voted correlator')
    case 2
        disp('Dual voted correlator')
    case 3
        disp(['Variable quantized correlator, ADC bits: ' num2str(ADC_bits)])
    otherwise
        disp('Undefined correlator.')
end
disp(['Sample rate: ' num2str(sample_rate)])
disp(['Error(s) detected at SNR: ' num2str(final_radio_snr) 'dB, with ' num2str(error_percent) ' % ADC errors.'])
disp(['Noise factor: ' num2str(final_noise_factor)])
disp(['Standard deviation of: SNR=' num2str(snr_std_deviation) ' Errors: ' num2str(error_counter_std_deviation) ' NoiseFactor: ' num2str(noise_factor_std_deviation)])
disp(['Calculation time: ' num2str(TimeSpent) ' seconds.'])


matlabpool('close');
% Turn on warnings again
warning('on', 'MATLAB:mir_warning_maybe_uninitialized_temporary');
