% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency, and correlated
% with a local PN code (Barker-11). The ADC resolution cannot be set yet.

% Enter various constants for the simulation
noise_factor = 0.9;             % Set the amplitude of the noise
sample_rate = 2;                % Multiple of the chip rate
%ADC_bits = 1;                  % Conversion resolution, this must set the +/- magnitude of the PN sequence
noise_resolution = 100;         % How many samples of noise for every sample of data
data = [1 1 -1 -1 -1 1 1 -1];   % Data sequence to transmitt
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];

% Generate sequences

% Create the reference code used in the correlator
% Expanded to match the sampling rate
% and max value of the ADC
% TODO number representation, +/-1. Should it be 1 and 0? And also 3 and 0?
% Algorithms used must also take this into account
%temp = barker_11 * ADC_peak;
pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, 11*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * length(data);
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Mix noise into the data stream, and read again using the ADC
% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);
% Create a vector of noise
noise = noise_factor*randn(1, length(tx_data)*noise_resolution);
% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data = tx_temp + noise;
radio_data = radio_data * (1- var(abs(radio_data)));
% rx_data is radio_data sampled at every x sample. Values above or below 0
% are trunicated to +/- 1.
rx_data = radio_data(1:noise_resolution:length(radio_data));
error_counter = 0;
for i = 1 : length(rx_data)
    % Limit the signal to the ADC resolution
    if rx_data(i) >= 0
        rx_data(i) = 1;
    elseif rx_data(i) < 0
        rx_data(i) = -1;
    end
    % Count how many errors are present
    if rx_data(i) ~= tx_data(i)
        error_counter = error_counter + 1;
    end
end

% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
hold off;
title('Received analog data from radio link with AGC')

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(rx_data, 'r');
hold off;
title(['Sampled data from radio link, error percent: ' num2str((error_counter / length(rx_data))*100) '%'])

% Sliding correlation of the received signal with the PN code. Should
% get nice positive and negative peaks for each bit.
figure(2)
% Auto correlate
r1_xx = xcorr(pn_code, pn_code);
L = (length(r1_xx)-1) / 2;
l = -L: L;
subplot(3, 1, 1);
plot(l, r1_xx);
title('Autocorrelation local PN code')
xlabel('Time displacement [samples]')

% Expected data to receive
subplot(3, 1, 2);
data_for_stem = zeros(1, (2 + length(data)));
data_for_stem(2:1+length(data)) = data;
data_for_stem(1) = 0;
%data_for_stem(end+1) = 0;
stem(data_for_stem);
title('Expected data sequence')

% Cross correlate
r1_xy = xcorr(rx_data, pn_code);
L = length(rx_data) + length(pn_code) - 1;
l = 0: L;
start_corr = length(r1_xy) - L;
subplot(3, 1, 3);
plot(l, r1_xy(start_corr:length(r1_xy)));
title('Cross Correlation of Received data and local PN code')
xlabel('Sample number')
