% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency, and correlated
% with a local PN code (Barker-11). The correlator is a matched filter
% (parallel correlator)

% Enter various constants for the simulation
noise_factor = 0.9;             % Set the amplitude of the noise
sample_rate = 4;                % Multiple of the chip rate
noise_resolution = 100;         % How many samples of noise for every sample of data
data = [1 1 -1 -1 -1 1 1 -1];   % Data sequence to transmitt
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];     % Needs sharp autocorrelation curve

% Generate sequences

% Create the reference code used in the correlator
% Expanded to match the sampling rate
pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, length(barker_11)*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * length(data);
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Mix noise into the data stream, and read again using the ADC
% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);
% Create a vector of noise
noise = noise_factor*randn(1, length(tx_data)*noise_resolution);
% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data = tx_temp + noise;
radio_snr = 10 * log10(mean(tx_temp.^2) / mean(noise.^2));
radio_data = radio_data * (1- var(abs(radio_data)));

% rx_data is radio_data sampled at every x sample.
rx_data = radio_data(1:noise_resolution:length(radio_data));
temp = rx_data;
% Next the data is quantized to allowable levels
error_counter = 0;
for i = 1 : length(rx_data)
    % Quantize each sample
    if rx_data(i) >= 0
        rx_data(i) = 1;
    else
        rx_data(i) = -1;
    end
    
    % Count how many errors are present
    if sign(rx_data(i)) ~= sign(tx_data(i))
        error_counter = error_counter + 1;
    end
end

% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
legend('Radio signal', 'TX Data');
hold off;
title(['Received analog data from radio link with AGC. SNR: ' num2str(radio_snr) ' dB'])

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(temp, 'g');
stem(rx_data, 'r');
legend('TX Data', 'Analog Sample', 'Digital Sample');
hold off;
title(['Sampled 1 bit data from radio link, error percent (sign): ' num2str((error_counter / length(rx_data))*100) '%'])

% Sliding correlation of the received signal with the PN code. Should
% get nice positive and negative peaks for each bit.
figure(2)

% Expected data to receive
subplot(2, 1, 1);
data_for_stem = zeros(1, (2 + length(data)));
data_for_stem(2:1+length(data)) = data;
data_for_stem(1) = 0;
stem(data_for_stem);
title('Expected data sequence')

% Cross correlate
% Matched filter algorithm
input_reg = zeros(1, length(pn_code));
%output = zeros(1, (length(rx_data) + length(pn_code) - 1));
output = zeros(1, length(rx_data));
for k = 1:length(rx_data)
    % Shift in new data
    input_reg(2:end) = input_reg(1:end-1);
    input_reg(1) = rx_data(k);
    
    % Multiply the register with the local PN code, and sum
    output(k) = input_reg * fliplr(pn_code)';
end

subplot(2, 1, 2);
plot(output);
title(['Cross Correlation of Received data and local PN code, sample rate ' num2str(sample_rate) ' times chip frequency.'])
xlabel('Sample number')
