% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency, and correlated
% with a local PN code (Barker-11). Signals are normalized to 1, ADC bits
% set how many discrete levels the sampled signal may have take.

% Enter various constants for the simulation
desired_snr = 0.2;               % In decibels
sample_rate = 8;                % Multiple of the chip rate
ADC_bits = 2;                   % Quantization levels, 2^ADC_bits
noise_resolution = 100;         % How many samples of noise for every sample of data
data = [1 1 -1 -1 -1 1 1 -1];   % Data sequence to transmitt
PN_Code = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];     % Barker-11 code

%% Generate sequences

% Create the reference code used in the correlator
% Expanded to match the sampling rate
pn_code = reshape((PN_Code' * ones(1, sample_rate))', 1, length(PN_Code)*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * length(data);
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Mix noise into the data stream, and read again using the ADC
% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);

% Create noise and scale to match desired SNR
noise = randn(1, length(tx_data)*noise_resolution);
noise = noise / norm(noise) * norm(tx_temp) / 10.0^(0.05*desired_snr);

% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data_noise = tx_temp + noise;
radio_data = radio_data_noise * (1- var(abs(radio_data_noise)));

% Determine SNR
radio_snr = 10 * log10(mean(tx_temp.^2) / mean(noise.^2));

%% rx_data is radio_data sampled at every x sample.
sampled_radio = radio_data(1:noise_resolution:length(radio_data));
% and quantized
rx_data = variable_bit_quantize( sampled_radio, ADC_bits );

error_counter = 0;
for i = 1 : length(rx_data)
    % Count how many errors are present
    if sign(rx_data(i)) ~= sign(tx_temp(i))
        error_counter = error_counter + 1;
    end
end

%% Correlate
% Autocorrelation of the PN code
r1_xx = xcorr(pn_code, pn_code);

% Cross correlation of radio data and PN code
r1_xy = xcorr(rx_data, pn_code);
L = length(rx_data) + length(pn_code) - 1;
start_corr = length(r1_xy) - L;
r1_xy = r1_xy(start_corr:end);

% Determine thresholds
thres_level = 2.8;
upper_thres = thres_level * sample_rate * ones(1, length(r1_xy));
lower_thres = -thres_level * sample_rate * ones(1, length(r1_xy));

%% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
legend('Radio signal', 'TX Data');
hold off;
title(['Received analog data from radio link with AGC. SNR: ' num2str(radio_snr) ' dB'])

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(sampled_radio, 'g');
stem(rx_data, 'r');
legend('TX Data', 'Analog Sample', 'Digital Sample');
hold off;
title(['Sampled ' num2str(ADC_bits) ' bit data from radio link, error percent (sign): ' num2str((error_counter / length(rx_data))*100) '%'])

% Sliding correlation of the received signal with the PN code. Should
% get nice positive and negative peaks for each bit.
figure(2)
% Auto correlate
L = (length(r1_xx)-1) / 2;
l = -L: L;
subplot(3, 1, 1);
plot(l, r1_xx);
title('Autocorrelation local PN code')
xlabel('Time displacement [samples]')

% Expected data to receive
subplot(3, 1, 2);
data_for_stem = zeros(1, (2 + length(data)));
data_for_stem(2:1+length(data)) = data;
data_for_stem(1) = 0;
stem(data_for_stem);
title('Expected data sequence')

% Cross correlate
subplot(3, 1, 3);
hold on;
plot(r1_xy);
plot(upper_thres, 'r');
plot(lower_thres, 'r');
hold off;
title(['Cross Correlation of Received data and local PN code, sample rate ' num2str(sample_rate) ' times chip frequency'])
xlabel('Sample number')
