% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency.

% Enter various constants for the simulation
sample_rate = 6;                % Multiple of the chip rate, depends on correlator design
desired_snr = 0.2;                % In decibels
data = [1 1 -1 -1 -1 1 1 -1 -1 -1 -1 1 1 1];   % Data sequence to transmitt
noise_resolution = 100;         % How many samples of noise for every sample of data

%% Generate sequences
barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];

% Create the reference code used in the correlator
% Expanded to match the sampling rate
pn_code = reshape((barker_11' * ones(1, sample_rate))', 1, length(barker_11)*sample_rate);

% Create the data sequence
% Expand the data to be sent with the normal or inverted PN code
tx_length = length(pn_code) * length(data);
tx_data = reshape((data' * pn_code)', 1, tx_length);

% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);

% Create noise and scale to match desired SNR
noise = randn(1, length(tx_data)*noise_resolution);
noise = noise / norm(noise) * norm(tx_temp) / 10.0^(0.05*desired_snr);

% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data_noise = tx_temp + noise;
radio_data = radio_data_noise * (1- var(abs(radio_data_noise)));
% Verify that the SNR is as desired
radio_snr = 10 * log10(mean(tx_temp.^2) / mean(noise.^2));

% rx_data is radio_data sampled at every x sample.
rx_data = radio_data(1:noise_resolution:length(radio_data));
temp = rx_data;
% Next the data is quantized to allowable levels
error_counter = 0;
for i = 1 : length(rx_data)
    % Quantize each sample
    if rx_data(i) >= 0
        rx_data(i) = 1;
    else
        rx_data(i) = 0;
    end
    
    % Count how many errors are present
    if (rx_data(i) ~= tx_data(i))
        if ~((tx_data(i) == -1) && (rx_data(i) == 0))
             error_counter = error_counter + 1;
        end
    end
end

%% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
legend('Radio signal', 'TX Data');
hold off;
title(['Received analog data from radio link with AGC. SNR: ' num2str(radio_snr) ' dB'])

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(temp, 'g');
stem(rx_data, 'r');
legend('TX Data', 'Analog Sample', 'Digital Sample');
hold off;
title(['Sampled 1 bit data from radio link, error percent: ' num2str((error_counter / length(rx_data))*100) '%'])

%% Write the received data vector to a file, for use in a testbench
str1 = ['Data sequence: [' num2str(data) ']'];
str2 = ['Signal SNR: ' num2str(radio_snr) 'dB. '];
str3 = ['Sampling rate: ' num2str(sample_rate)];
str4 = 'Received data vector from 1-bit ADC:';

fName = 'rx_vect_for_tb.txt';
fid = fopen(fName,'w');
if fid ~= -1
  fprintf(fid,'%s\r\n',str1);
  fprintf(fid,'%s\r\n',str2); 
  fprintf(fid,'%s\r\n',str3);
  fprintf(fid,'%s\r\n',str4);
  fclose(fid);                   
end
dlmwrite(fName, rx_data', '-append',... 
         'delimiter','\t',...
         'newline','pc');
