//-----------------------------------------------------------------------------
//
// Title       : dual_corr
// Author      : Eirik Taylor, 2014
//
//-----------------------------------------------------------------------------
//
// Description : This correlator combines the output of several smaller correlators
//				which have been multiplexed. Result is one ADC sample delayed,
//				quite possibly only due to the mini correlators used.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module dual_corr (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam mini_correlators = 2;
	localparam outputbits = 6;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define internal connections 	
	wire [outputbits-1:0] corr;
	reg current_correlator;
	
	// Connections to the mini correlators
	wire [4:0] corr_res_1, corr_res_2; 
	reg adc_rdy_1, adc_rdy_2;	 
	
	// Connect a mini correlator
	sub_corr_alt minicorr_one (
		.reset(reset),
		.clk(clk),
		.sample(sample),
		.corr(corr_res_1),
		.adc_ready(adc_rdy_1) 
		);
	
	// Connect a mini correlator
	sub_corr_alt minicorr_two (
		.reset(reset),
		.clk(clk),
		.sample(sample),
		.corr(corr_res_2),
		.adc_ready(adc_rdy_2) 
		);	
	
	// The total correlation is simply the sum of each sub correlator
	assign corr = corr_res_1 + corr_res_2;	
	
	// ADC action, clock a new value into the shift register, also update the state machine state 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					current_correlator <= 0;
					adc_rdy_1 <= 0;	   
					adc_rdy_2 <= 0;	
			end else begin
					if (adc_ready) begin
							// Shift the new sample in one of the mini correlators
							// Alternate between correlators
							if (current_correlator) begin 
									current_correlator <= 0;
									adc_rdy_1 <= 1;	
									adc_rdy_2 <= 0;
							end else begin	
									current_correlator <= 1;
									adc_rdy_1 <= 0;
									adc_rdy_2 <= 1;			 
								end
							
					end	else begin	 
							// Holde the correlator inputs idle
							adc_rdy_1 <= 0;	   
							adc_rdy_2 <= 0;	   
						end
				end
		end
	
endmodule
