//-----------------------------------------------------------------------------
//
// Title       : par_corr_mux
// Author      : Eirik Taylor, 2014
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses an 11-bit Barker code. ADC sampling rate assumed to be twice the chip rate.
//				Output result is ready some time before the next sample is expected.
//
//				Works by multing the input ADC samples with the barker code, and then summing
//				groups of 11-bits, by muxing a single 11-bit parallel counter across the mult
//				results.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

// TODO Implement sampling rate as a parameter
module par_corr_mux (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 4;		// NB! State machine must be updated manually!
	localparam outputbits = 6;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 11-bit Barker code	
	localparam barker11 = 11'b 11100010010;
	
	// Define internal connections 
	reg [outputbits-1:0] corr;
	reg [(11*oversamplingrate)-1:0] adc_samples;
	wire [(11*oversamplingrate)-1:0] mult_result;
	
	// Sum the output of every sample multiplication to get the correlation measure.
	reg [10:0] counted_section;
	wire [3:0] bitsum; 
	
	// Connect the parallel counter
	eleven_bit_cnter bitcounter (
		.in(counted_section),
		.sum(bitsum) 
		);
	
	// State machine stuff
	localparam IDLE=0, START=1, WAIT=2; 
	reg [2:0] state, nxtState; 
	
	// Multiply every element from the ADC register with an element in the Barker-11 code.	
	genvar i, j;
	generate
		// For every Barker bit
		for(i=0; i<11; i=i+1)
			begin : mult 
				// For every oversampled bit
				for(j=0; j<oversamplingrate; j=j+1)
					begin
						// This is a 1-bit system, so mult is done with an XNOR gate
						//xnor(mult_result[j + i*2], adc_samples[j + i*2],  barker11[i]);	
						
						// This is a 1-bit system with known coefficients, so mult can be done with a inverter/buffer 
						if (barker11[i]) begin
								// Buffer
								assign mult_result[j + i*oversamplingrate] = adc_samples[j + i*oversamplingrate];
						end else begin
								// Inverter
								not(mult_result[j + i*oversamplingrate], adc_samples[j + i*oversamplingrate]);
							end
					end
			end
	endgenerate
	
	// ADC action, clock a new value into the shift register, also update the state machine state 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					corr <= 0;
					adc_samples <= 0; 
					state <= IDLE;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples <= {adc_samples[(11*oversamplingrate)-2:0], sample};
							// Reset the state machine
							corr <= 0; 
							state <= WAIT;
					end else begin
							// Change state machine state	
							state <= nxtState;
							// Sum the correlation result
							corr <= corr + bitsum; 
						end				
				end
		end	
	
	// State machine logic, select the input to the parallel counter
	always @(state) begin	  
			nxtState = IDLE;
			case (state)
				IDLE : begin
						counted_section = 0;
						nxtState = IDLE;
					end
				WAIT : begin
						// Wait a clock pulse so the ADC register can shift
						counted_section = 0;
						nxtState = START;
					end
				START : begin
						counted_section = mult_result[10:0];
						nxtState = 3;
					end
				3 : begin
						counted_section = mult_result[21:11];
						nxtState = 4;
					end
				4 : begin
						counted_section = mult_result[32:22];
						nxtState = 5;
					end 
				5 : begin
						counted_section = mult_result[43:33];
						nxtState = IDLE;
					end
				default : begin
						counted_section = 0;
						nxtState = IDLE;
					end
			endcase 
		end
	
endmodule
