//-----------------------------------------------------------------------------
//
// Title       : par_corr_ref
// Design      : parallel_correlator
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses an 11-bit Barker code. ADC sampling rate assumed to be twice the chip rate.
//				Output result is ready some time before the next sample is expected.
//
//				Works by multing the input ADC samples with the barker code, and then summing
//				groups of 11-bits.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

// TODO Implement sampling rate as a parameter
module par_corr_ref (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 4;		// NB! Summing function must be created manually
	localparam outputbits = 6;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 11-bit Barker code	
	localparam barker11 = 11'b 11100010010;
	
	// Define internal connections 
	wire [outputbits-1:0] corr;
	reg [(11*oversamplingrate)-1:0] adc_samples;
	wire [(11*oversamplingrate)-1:0] mult_result;
	
	// Sum the output of every sample multiplication to get the correlation measure.
	wire [3:0] bitsum [0:oversamplingrate-1]; 
	
	// Connect the required number of 11-bit parallel counters
	genvar k;
	generate
		for(k=0; k<oversamplingrate; k=k+1)
			begin : parallel_add 
				// Connect the parallel counter
				eleven_bit_cnter bitcounter (
				.in(mult_result[(k*11)+10 : k*11]),
				.sum(bitsum[k][3:0]) 
				);
			end
	endgenerate	 
	
	// NB! Summing function isn't generated automatically
	// Find the sum
	assign corr = bitsum[0] + bitsum[1] + bitsum[2] + bitsum[3];
	
	// Multiply every element from the ADC register with an element in the Barker-11 code.	
	genvar i, j;
	generate
		// For every Barker bit
		for(i=0; i<11; i=i+1)
			begin : mult 
				// For every oversampled bit
				for(j=0; j<oversamplingrate; j=j+1)
					begin
						// This is a 1-bit system, so mult is done with an XNOR gate
						//xnor(mult_result[j + i*2], adc_samples[j + i*2],  barker11[i]);	
						
						// This is a 1-bit system with known coefficients, so mult can be done with a inverter/buffer 
						if (barker11[i]) begin
								// Buffer
								assign mult_result[j + i*oversamplingrate] = adc_samples[j + i*oversamplingrate];
						end else begin
								// Inverter
								not(mult_result[j + i*oversamplingrate], adc_samples[j + i*oversamplingrate]);
							end
					end
			end
	endgenerate
	
	// ADC action, clock a new value into the shift register 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin 
					adc_samples <= 0;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples <= {adc_samples[(11*oversamplingrate)-2:0], sample};
						end
				end
		end
	
endmodule
