//-----------------------------------------------------------------------------
//
// Title       : three_bit_voted
// Author      : Eirik Taylor, 2014
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses an 11-bit Barker code. Output result is ready some time before the next sample is expected.
//
//				Works by multing the input ADC samples with the barker code, and then voting
//				on groups of samples, before counting them using an up-counter. The result from the counter
//				is latched, to prevent spurious output while counting. Output is latched on the next ADC clock.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module three_bit_voted (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 3;		// NB! Determined by the voter size, must be odd.
	localparam outputbits = 4;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 11-bit Barker code	
	localparam barker11 = 11'b 11100010010;
	
	// Define internal connections 
	reg [outputbits-1:0] corr;
	reg [(11*oversamplingrate)-1:0] adc_samples;
	wire [(11*oversamplingrate)-1:0] mult_result;
	
	// Connections to the counter and voter
	reg [2:0] voted_section;
	wire voted_bit;
	reg counter_ctrl;
	reg reset_count;
	wire [3:0] counter_value;
	
	// Connect the upcounter
	upcounter counter (
		.count(counter_value),	
		.clk(clk),
		.up(counter_ctrl),
		.reset(reset_count) 
		);
	
	// Connect the voter unit
	voter_unit voter (
		.ballots(voted_section),
		.vote(voted_bit) 
		);
	
	// State machine stuff
	localparam IDLE=0, START=1, WAIT=2; 
	reg [3:0] state, nxtState; 
	
	// Multiply every element from the ADC register with an element in the Barker-11 code.	
	genvar i, j;
	generate
		// For every Barker bit
		for(i=0; i<11; i=i+1)
			begin : mult 
				// For every oversampled bit
				for(j=0; j<oversamplingrate; j=j+1)
					begin
						// This is a 1-bit system, so mult is done with an XNOR gate
						//xnor(mult_result[j + i*2], adc_samples[j + i*2],  barker11[i]);	
						
						// This is a 1-bit system with known coefficients, so mult can be done with a inverter/buffer 
						if (barker11[i]) begin
								// Buffer
								assign mult_result[j + i*oversamplingrate] = adc_samples[j + i*oversamplingrate];
						end else begin
								// Inverter
								not(mult_result[j + i*oversamplingrate], adc_samples[j + i*oversamplingrate]);
							end
					end
			end
	endgenerate
	
	// ADC action, clock a new value into the shift register, also update the state machine state 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					reset_count <= 0;
					counter_ctrl <= 0;
					adc_samples <= 0; 
					corr <= 0;
					state <= IDLE;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples <= {adc_samples[(11*oversamplingrate)-2:0], sample};
							// Update the output
							corr <= counter_value;
							// Reset the state machine and counter
							reset_count <= 0;
							counter_ctrl <= 0;
							state <= WAIT;
					end else begin
							// Change state machine state	
							state <= nxtState;
							reset_count <= 1;
							counter_ctrl <= voted_bit;
						end				
				end
		end	
	
	// State machine logic, select the input to the parallel counter
	always @(state) begin	  
			nxtState = IDLE;
			case (state)
				IDLE : begin
						voted_section = 0;
						nxtState = IDLE;
					end
				WAIT : begin
						// Wait a clock pulse so the ADC register can shift
						voted_section = 0;
						nxtState = START;
					end
				START : begin
						voted_section = mult_result[2:0];
						nxtState = 3;
					end
				3 : begin
						voted_section = mult_result[5:3];
						nxtState = 4;
					end
				4 : begin
						voted_section = mult_result[8:6];
						nxtState = 5;
					end 
				5 : begin
						voted_section = mult_result[11:9];
						nxtState = 6;
					end
				6 : begin
						voted_section = mult_result[14:12];
						nxtState = 7;
					end
				7 : begin
						voted_section = mult_result[17:15];
						nxtState = 8;
					end
				8 : begin
						voted_section = mult_result[20:18];
						nxtState = 9;
					end
				9 : begin
						voted_section = mult_result[23:21];
						nxtState = 10;
					end
				10 : begin
						voted_section = mult_result[26:24];
						nxtState = 11;
					end
				11 : begin
						voted_section = mult_result[29:27];
						nxtState = 12;
					end
				12 : begin
						voted_section = mult_result[32:30];
						nxtState = IDLE;
					end
				default : begin
						voted_section = 0;
						nxtState = IDLE;
					end
			endcase 
		end
	
endmodule
