//-----------------------------------------------------------------------------
//
// Title       : sub_corr
// Author      : Eirik Taylor, 2014
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses an 11-bit Barker code. Output result is ready when the next ADC
//				sample has arrived.
//
//				Set oversampling rate of 2. Output needs to be latched between calculations.
//
//				Works by multing the input ADC samples with the barker code, and then shifting the result
//				through a 5-bit counter. Can the shifting can be done with the ADC register somehow?
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module sub_corr (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 2;
	localparam outputbits = 5;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 11-bit Barker code	
	localparam barker11 = 11'b 11100010010;
	
	// Define internal connections 
	reg [outputbits-1:0] corr;
	reg [(11*oversamplingrate)-1:0] adc_samples;
	wire [(11*oversamplingrate)-1:0] mult_result;
	
	// Connections to the counter and voter
	reg [(11*oversamplingrate)-1:0] mult_res_shifter; 
	reg reset_count;
	wire [outputbits-1:0] counter_value;   
	reg [1:0] pulse_delay;
	
	// Connect the upcounter
	upcounter counter (
		.count(counter_value),	
		.clk(clk),
		.up(mult_res_shifter[(11*oversamplingrate)-1]),
		.reset(reset_count) 
		);
	
	// Multiply every element from the ADC register with an element in the Barker-11 code.	
	genvar i, j;
	generate
		// For every Barker bit
		for(i=0; i<11; i=i+1)
			begin : mult 
				// For every oversampled bit
				for(j=0; j<oversamplingrate; j=j+1)
					begin
						// This is a 1-bit system, so mult is done with an XNOR gate
						//xnor(mult_result[j + i*2], adc_samples[j + i*2],  barker11[i]);	
						
						// This is a 1-bit system with known coefficients, so mult can be done with a inverter/buffer 
						if (barker11[i]) begin
								// Buffer
								assign mult_result[j + i*oversamplingrate] = adc_samples[j + i*oversamplingrate];
						end else begin
								// Inverter
								not(mult_result[j + i*oversamplingrate], adc_samples[j + i*oversamplingrate]);
							end
					end
			end
	endgenerate
	
	// ADC action, clock a new value into the shift register, also update the state machine state 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					adc_samples <= 0;
					corr <= 0;
					reset_count <= 0;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples <= {adc_samples[(11*oversamplingrate)-2:0], sample};  
							// Latch the output
							corr <= counter_value;
							reset_count <= 0;
					end	else begin
							// Reset needs to be controlled sequentially here, otherwise it is reset before the output is latched
							reset_count <= 1;
						end
				end
		end	 
	
	// The mult result register, used by the counter
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					mult_res_shifter <= 0;
			end else begin 
					// The shift register is loaded two clocks after ADC ready goes high
					if (pulse_delay[1]) begin	  
							// Load the reg
							mult_res_shifter <= mult_result;
					end else begin
							// Shift the reg 
							mult_res_shifter <= {mult_res_shifter[(11*oversamplingrate)-2:0], 1'b0};
						end
				end
		end	 
	
	// Clock delay used to control loading of the mult result register
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					pulse_delay <= 0;
			end else begin 
					// Shift the register
					pulse_delay <= {pulse_delay[0], adc_ready};
				end
		end
	
endmodule
