//-----------------------------------------------------------------------------
//
// Title       : three_bit_voted_alt
// Author      : Eirik Taylor, 2014
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses an 11-bit Barker code. Output result is ready some time before the next sample is expected.
//
//				Works by voting on groups of input ADC samples, and then multiplying with the
//				barker code	before counting them using an 11-bit parallel counter.	 
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module three_bit_voted_alt (reset, clk, sample, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 3;		// NB! Determined by the voter size, must be odd.
	localparam outputbits = 4;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 11-bit Barker code	
	localparam barker11 = 11'b 11100010010;	
	
	// Define internal connections 
	wire [outputbits-1:0] corr;
	reg [(11*oversamplingrate)-1:0] adc_samples;
	
	// Connections to the counter and voter
	wire [10:0] voted_results; 
	wire [10:0] mult_result;
	
	// Connect the upcounter
	eleven_bit_cnter bitcounter (
		.in(mult_result),
		.sum(corr) 
		);
	
	// Connect the required number of voter units
	genvar k;
	generate
		for(k=0; k<11; k=k+1)
			begin : oversample_voter 
				// Connect the 3-bit voter to each section of oversampled bits
				voter_unit voter (
				.ballots(adc_samples[(oversamplingrate-1)+oversamplingrate*k:oversamplingrate*k]),
				.vote(voted_results[k]) 
				);
			end
	endgenerate	
	
	// Multiply each voted bit with an element in the Barker-11 code.	
	genvar i, j;
	generate
		// For every Barker bit
		for(i=0; i<11; i=i+1)
			begin : mult 
				// This is a 1-bit system with known coefficients, so mult can be done with a inverter/buffer 
				if (barker11[i]) begin
						// Buffer
						assign mult_result[i] = voted_results[i];
				end else begin
						// Inverter
						not(mult_result[i], voted_results[i]);
					end
			end
	endgenerate
	
	// ADC action, clock a new value into the shift register, also update the state machine state 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin
					adc_samples <= 0; 
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples <= {adc_samples[(11*oversamplingrate)-2:0], sample};
						end		
				end
		end	
	
endmodule
