% Generate a radio signal with a random data set, while increasing the noise
% level. Noise if first increased in large steps, until an error is
% detected. Then the algorithm jumps back two steps, and repeats with a
% smaller step size. The result once an error is detected now is stored.
% The process is repeated a number of times, and the final average is
% taken.

%%
% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency.

% Enter various constants for the simulation
sample_rate = 3;                % Multiple of the chip rate, depends on correlator design
correlator_type = 4;            % 0 = standard, 1 = voted, 2 = dual voted, 3 = variable ADC, 4 = 2-bit standard, 5 = 2-bit voted
ADC_bits = 3;                   % Only used for correlator_type = 3
noise_increment_fine = 0.005;         % Set the fine simulation resolution
noise_increment_rough = 0.1;         % Set the rough simulation resolution
averaging_passes = 20;
data_vector_length = 1000;

%% Generate sequences
Barker_11 = [+1 +1 +1 -1 -1 -1 +1 -1 -1 +1 -1];
chirp_36 = [ 1 zeros(1,7) -1 zeros(1,6) 1 zeros(1,5) -1 zeros(1,4) 1 zeros(1,3) -1 zeros(1,2) 1 0 -1 ];

% Pick the correct PN code for the correlator model
if (correlator_type >= 4)
    pn_code = chirp_36;
else
    pn_code = Barker_11;
end
pn_code_len = length(pn_code);

% Vector to hold each result
SNR_vector = zeros(1, averaging_passes);
error_counter_vector = zeros(1, averaging_passes);
noise_factor_vector = zeros(1, averaging_passes);

% Open a worker pool for parallel computation
matlabpool('open');

poolSize = matlabpool('size');
if poolSize == 0
   error('parallel:demo:poolClosed', ...
        'This demo needs an open MATLAB pool to run.');
end
fprintf('This demo is running on %d MATLABPOOL workers.\n', ...
    matlabpool('size'));
% Turn off temporary variable warnings, temporarily
warning('off', 'MATLAB:mir_warning_maybe_uninitialized_temporary');

%time the calculation
tic;

% Run the algorithm a number of times, and average the result
parfor q = 1:averaging_passes
    % Starting point for the noise level
    noise_factor = 0.0;             % Set the start amplitude of the noise
    noise_end = 1.5;                % Stopping point for the simulation
    % Start the simulation with rough increments, to save time
    noise_increment = noise_increment_rough;
    errors = 0;
    
    while((errors == 0) && (noise_factor < noise_end))
        % Generate a random vector of 1 and -1, representing data to be used at
        % this noise level
        data = 2*randi(2, 1, data_vector_length) - 3;

        % Create the reference code used in the correlator
        % Expanded to match the sampling rate
        pn_code_expanded = reshape((pn_code' * ones(1, sample_rate))', 1, length(pn_code)*sample_rate);
        
        % Create the data sequence
        % Expand the data to be sent with the normal or inverted PN code
        tx_length = length(pn_code_expanded) * data_vector_length;
        tx_data = reshape((data' * pn_code_expanded)', 1, tx_length);

        % Mix noise into the data stream, and read again using the ADC
        % Create a vector of noise
        % Increase the noise with each pass
        noise_factor = noise_factor + noise_increment;
        noise = noise_factor*randn(1, length(tx_data));
        % Noise is assumed to be added to the signal. The receiver will likely have
        % an Automatic-Gain-Control stage, normalizing the signal amplitude
        radio_data = tx_data + noise;
        radio_snr = 10 * log10(mean(tx_data.^2) / mean(noise.^2));
        radio_data = radio_data * (1- var(abs(radio_data)));

        %% Cross correlate
        % Select the correlator and thresholds to use

        switch correlator_type
           case 0
               %% Standard correlator

                % Quantized to 1-bit
                rx_data = one_bit_quantize(radio_data);
                % Correlate
                correlation_vector = standard_correlation_1bit(rx_data, sample_rate);

                % Thresholds have been tuned
                lower_thres = floor(3 * sample_rate);
                upper_thres = (sample_rate * 11) - lower_thres;

           case 1
              %% Voted type
                if mod(sample_rate,2) == 0
                   error('Sampling rate is even! Use an odd sampling rate for the voted unit.')
                end

                % Quantized to 1-bit
                rx_data = one_bit_quantize(radio_data);
                % Correlate
                correlation_vector = voted_correlator_1bit(rx_data, sample_rate);

                % Thresholds have been tuned
                upper_thres = 9;
                lower_thres = 2;

           case 2
              %% Dual voted type
                % Check if the sampling rates align
                sample_rate_mini = 3;           % The sampling rate of the sub correlator, for mode 2
                if rem(sample_rate, sample_rate_mini) 
                    error('Sampling rate must be multiple of sub correlator rate!')
                else
                    mini_corr_units = sample_rate / sample_rate_mini;   % How many mini corrs
                end

                % Quantized to 1-bit
                rx_data = one_bit_quantize(radio_data);
                % Correlate
                correlation_vector = dual_voted_correlation_1bit(rx_data, sample_rate);

                % Thresholds have been tuned
                upper_thres = 18;
                lower_thres = 4;

            case 3
                %% Variable bit quantized
                rx_data = variable_bit_quantize( radio_data, ADC_bits );

                % Cross correlation of radio data and PN code
                correlation_vector = xcorr(rx_data, pn_code_expanded);
                L = length(rx_data) + length(pn_code_expanded) - 1;
                start_corr = length(correlation_vector) - L;
                correlation_vector = correlation_vector(start_corr:end);

                % Determine thresholds
                thres_level = 2.8;
                upper_thres = thres_level * sample_rate;
                lower_thres = -thres_level * sample_rate;

            case 4
                %% 2-bit quantized, standard output
                rx_data = two_thres_quantize(radio_data);
                % Correlate
                correlation_vector = offsetbin_standard_correlation_2bit(rx_data, sample_rate);
                % Thresholds have been tuned
                midpoint = (16/2) * sample_rate;
                thres = 4.4 * sample_rate;
                upper_thres = midpoint + thres;
                lower_thres = midpoint - thres;
                
            case 5
                %% 2-bit quantized, voted output
                rx_data = two_thres_quantize(radio_data);
                % Correlate
                correlation_vector = voted_correlation_2bit(rx_data, sample_rate);
                % Thresholds have been tuned
                midpoint = 0;
                thres = 4.4;
                upper_thres = midpoint + thres;
                lower_thres = midpoint - thres;

            otherwise
            error('Error! Undefined correlator used!')

        end

        % Count sign errors between the received data and the sent data
        % Rough measure of the nosie impact
        error_counter = 0;
        for i = 1 : length(rx_data)
            % Count how many errors are present
            if sign(rx_data(i)) ~= sign(tx_data(i))
                error_counter = error_counter + 1;
            end
        end

        %% Detect data bits in the correlator vector
        detected_data = detect_correlator_peaks(correlation_vector, pn_code_len, upper_thres, lower_thres, sample_rate);

        % Count errors
        errors = 0;
        if (length(detected_data) > length(data))
            for l = 1:length(data)
                if (detected_data(l) ~= data(l))
                    errors = errors + 1;
                end
            end
        else
            for l = 1:length(detected_data)
                if (detected_data(l) ~= data(l))
                    errors = errors + 1;
                end
            end
        end

        % Check if this was the first pass, if so use finer resolution and try
        % again
        if ((errors ~= 0) && (noise_increment == noise_increment_rough))
            noise_factor = noise_factor - 2*noise_increment_rough;
            errors = 0;
            noise_increment = noise_increment_fine;
        end

    end
    
    % Store results from the pass
    SNR_vector(q) = radio_snr;
    error_counter_vector(q) = error_counter;
    noise_factor_vector(q) = noise_factor;
    
    % Keep the user awake
    disp(['Pass ' num2str(q) '. SNR: ' num2str(radio_snr) 'dB and noise factor: ' num2str(noise_factor)])
end

TimeSpent = toc;

% Calculate the averages
final_radio_snr = mean(SNR_vector);
final_error_counter = mean(error_counter_vector);
final_noise_factor = mean(noise_factor_vector);

% Calculate standard deviation of each parameter
snr_std_deviation = std(SNR_vector);
error_counter_std_deviation = std(error_counter_vector);
noise_factor_std_deviation = std(noise_factor_vector);

%% Display results
error_percent = 100 * (final_error_counter / (length(pn_code) * sample_rate * data_vector_length));
error_std_percent = 100 * (error_counter_std_deviation / final_error_counter);
switch correlator_type
    case 0
        disp('Standard 1-bit correlator')
	case 1
        disp('Voted 1-bit correlator')
    case 2
        disp('Dual voted 1-bit correlator')
    case 3
        disp(['Variable quantized correlator, ADC bits: ' num2str(ADC_bits)])
    case 4
        disp('Standard 2-bit correlator')
    case 5
        disp('Voted 2-bit correlator')
    otherwise
        disp('Undefined correlator.')
end
disp(['Sample rate: ' num2str(sample_rate)])
disp(['Error(s) detected at SNR: ' num2str(final_radio_snr) 'dB, with ' num2str(error_percent) ' % ADC sign errors.'])
disp(['Noise factor: ' num2str(final_noise_factor)])
disp(['Standard deviation of: SNR=' num2str(snr_std_deviation) ' Errors: ' num2str(error_std_percent) '% NoiseFactor: ' num2str(noise_factor_std_deviation)])
disp(['Calculation time: ' num2str(TimeSpent) ' seconds.'])


matlabpool('close');
% Turn on warnings again
warning('on', 'MATLAB:mir_warning_maybe_uninitialized_temporary');
