% Generates a simulated radio transmission of a Direct sequence spread
% spectrum radio signal, with added white noise. The received signal is
% sampled by an ADC at a multiple of the chip frequency.

% Enter various constants for the simulation
sample_rate = 5;                % Multiple of the chip rate, depends on correlator design
desired_snr = 100;                % In decibels
data = [1 1 -1 -1 -1 1 1 -1 -1 -1 -1 1 1 1];   % Data sequence to transmitt
noise_resolution = 100;         % How many samples of noise for every sample of data

%% Generate sequences
PN_Code = [ 1 zeros(1,7) -1 zeros(1,6) 1 zeros(1,5) -1 zeros(1,4) 1 zeros(1,3) -1 zeros(1,2) 1 0 -1 ];

% Create the reference code used in the correlator
% Expand the data and PN code to match the sampling rate
pn_code = reshape((PN_Code' * ones(1, sample_rate))', 1, length(PN_Code)*sample_rate);
tx_data = reshape((data' * pn_code)', 1, length(pn_code) * length(data));

% Expand the tx_data so additional noise samples can be added
tx_temp = reshape((tx_data' * ones(1, noise_resolution))', 1, length(tx_data)*noise_resolution);

% Create noise and scale to match desired SNR
noise = randn(1, length(tx_data)*noise_resolution);
noise = noise / norm(noise) * norm(tx_temp) / 10.0^(0.05*desired_snr);

% Noise is assumed to be added to the signal. The receiver will likely have
% an Automatic-Gain-Control stage, normalizing the signal amplitude
radio_data_noise = tx_temp + noise;
radio_data = radio_data_noise * (1- var(abs(radio_data_noise)));
% Verify that the SNR is as desired
radio_snr = 10 * log10(mean(tx_temp.^2) / mean(noise.^2));

% rx_data is radio_data sampled at every x sample.
rx_data = radio_data(1:noise_resolution:length(radio_data));
temp = rx_data;
% Next the data is quantized to allowable levels
% adc_vector = zeros(2, length(rx_data));
% for i = 1 : length(rx_data)
%     % Quantize each sample
%     if (rx_data(i) >= 0.5)
%         adc_vector(1, i) = 1;
%         adc_vector(2, i) = 0;
%     elseif (rx_data(i) <= -0.5)
%         adc_vector(1, i) = 0;
%         adc_vector(2, i) = 1;
%     else
%         adc_vector(1, i) = 0;
%         adc_vector(2, i) = 0;
%     end
% end

adc_vector = two_thres_quantize(rx_data);

%% Plot the clean sampled data sequence, the data sequence with noise, and
% the resulting sampled data sequence. Display some error metric somewhere
% as well
figure(1)
% Plot received data with noise overlaid
subplot(2, 1, 1);
hold on;
plot(radio_data, 'b');
plot(tx_temp, 'r');
legend('Radio signal', 'TX Data');
hold off;
title(['Received analog data from radio link with AGC. SNR: ' num2str(radio_snr) ' dB'])

% Plot sampled data and expected data
subplot(2, 1, 2);
hold on;
stem(tx_data, 'b');
stem(rx_data, 'g');
stem(adc_vector(1, :), 'r');
stem(-adc_vector(2, :), 'r');
legend('TX Data', 'Analog Sample', 'Digital Sample');
hold off;
title('Sampled 2 bit data from radio link')

%% Write the received data vector to a file, for use in a testbench
str1 = ['Data sequence: [' num2str(data) ']'];
str2 = ['Signal SNR: ' num2str(radio_snr) 'dB. '];
str3 = ['Sampling rate: ' num2str(sample_rate)];
str4 = 'Received data vector from dual threshold ADC. Column 1 is positive sample, column 2 negative:';

fName = 'rx_vect_for_tb.txt';
fid = fopen(fName,'w');
if fid ~= -1
  fprintf(fid,'%s\r\n',str1);
  fprintf(fid,'%s\r\n',str2); 
  fprintf(fid,'%s\r\n',str3);
  fprintf(fid,'%s\r\n',str4);
  fclose(fid);                   
end
dlmwrite(fName, adc_vector', '-append',... 
         'delimiter','\t',...
         'newline','pc');
