//-----------------------------------------------------------------------------
//
// Title       : twobit_corr_addmux_sixteen
// Design      : parallel_correlator
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses a 36-bit chip as the PN code. ADC sampling rate can be set.
//				Output result is ready (1 + N) clock cycles after an ADC sample has arrived.
//				
//				Instead of muxing a single parallel counter to sections of the ADC register,
//				the full number of counters is used. Rather, the result from each counter is
//				muxed to the integrator.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module twobit_corr_addmux_sixteen (reset, clk, sample_pos, sample_neg, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 5;		// NB! The state machine must be scaled after changing this!
	localparam outputbits = 7;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample_pos;
	input sample_neg;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 36-bit chirp, split into positive and negative sections
	//localparam PN_Code_pos = 36'b 100000000000000100000000001000000100;
	//localparam PN_Code_neg = 36'b 000000001000000000000100000000100001;	
	
	// Define internal connections 
	reg [outputbits-1:0] corr;
	reg [(36*oversamplingrate)-1:0] adc_samples_pos;	
	reg [(36*oversamplingrate)-1:0] adc_samples_neg;
	
	// Sum the output of every sample multiplication to get the correlation measure.  
	wire [4:0] bitsum [0:oversamplingrate-1];
	reg [4:0] partial_sum;
	
	// Used for the inverse of the PN Code
	wire [3:0] positive_zeros [0:oversamplingrate-1]; 
	wire [3:0] negative_zeros [0:oversamplingrate-1];  
	
	// State machine stuff
	localparam IDLE=0, START=1, WAIT=2; 
	reg [2:0] state, nxtState;
	
	// Connect the required number of 4-bit parallel counters to the adc registers	
	// One for each code, and one for the "inverse" of each code, times the sample rate
	// So 4 * sample rate
	genvar k;
	generate
		for(k=0; k<oversamplingrate; k=k+1)
			begin : parallel_add
				
				// Connect the parallel counter to the positive samples, where there must be zeros
				not(positive_zeros[k][0], adc_samples_pos[k]);
				not(positive_zeros[k][1], adc_samples_pos[5*oversamplingrate+k]);
				not(positive_zeros[k][2], adc_samples_pos[14*oversamplingrate+k]);
				not(positive_zeros[k][3], adc_samples_pos[27*oversamplingrate+k]); 
				
				// Connect the parallel counter to the negative samples, where there must be zeros
				not(negative_zeros[k][0], adc_samples_neg[2*oversamplingrate+k]);
				not(negative_zeros[k][1], adc_samples_neg[9*oversamplingrate+k]);
				not(negative_zeros[k][2], adc_samples_neg[20*oversamplingrate+k]);
				not(negative_zeros[k][3], adc_samples_neg[35*oversamplingrate+k]);
				
				// Connect the parallel counter to the positive samples and negative samples
				sixteen_bit_cnter bitcounter_pos (
				.in({adc_samples_pos[2*oversamplingrate+k], adc_samples_pos[9*oversamplingrate+k], adc_samples_pos[20*oversamplingrate+k], adc_samples_pos[35*oversamplingrate+k], positive_zeros[k], adc_samples_neg[k], adc_samples_neg[5*oversamplingrate+k], adc_samples_neg[14*oversamplingrate+k], adc_samples_neg[27*oversamplingrate+k], negative_zeros[k]}),
				.sum(bitsum[k][4:0]) 
				);
				
			end
	endgenerate
	
	// ADC action, clock a new value into the shift registers 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin  
					corr <= 0;
					state <= IDLE;
					adc_samples_pos <= 0;  
					adc_samples_neg <= 0;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples_pos <= {adc_samples_pos[(36*oversamplingrate)-2:0], sample_pos};	
							adc_samples_neg <= {adc_samples_neg[(36*oversamplingrate)-2:0], sample_neg};   
							// Reset the state machine
							corr <= 0; 
							state <= WAIT;
					end else begin
							// Change state machine state	
							state <= nxtState;
							corr <= corr + partial_sum; 
						end
				end
		end	 
	
	// State machine used to multiplex the four bit parallel counter results to the integrator
	always @(state) begin	  
			nxtState = IDLE;
			case (state)
				IDLE : begin
						partial_sum = 0;
						nxtState = IDLE;
					end
				WAIT : begin
						// Wait a clock pulse so the ADC register can shift
						partial_sum = 0;
						nxtState = START;
					end
				START : begin
						partial_sum = bitsum[0];
						nxtState = 3;
					end
				3 : begin
						partial_sum = bitsum[1];
						nxtState = 4;
					end
				4 : begin
						partial_sum = bitsum[2];
						nxtState = 5;
					end
				5 : begin
						partial_sum = bitsum[3];
						nxtState = 6;
					end
				6 : begin
						partial_sum = bitsum[4];
						nxtState = IDLE;
					end
				default : begin
						partial_sum = 0;
						nxtState = IDLE;
					end
			endcase 
		end
	
endmodule
