//-----------------------------------------------------------------------------
//
// Title       : hash_corr_three_alt
// Design      : hash_corr
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Correlator for Direct Sequence Spread Spectrum radio.
//				Uses a 36-bit chip as the PN code.
//
//				Uses a 5-bit hash to store sample groups of three, which otherwise require
//				6-bits. A hash generator must run through each 5-bit hash to update them,
//				for every new sample recevied. 
//
//				Alternate by not using a case selector, but boolean expressions
//				instead for the count decoders.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module hash_corr_three_alt (reset, clk, sample_pos, sample_neg, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 3;		// NB! Must be 3
	localparam outputbits = 6;
	
	//localparam register_size = ((36*oversamplingrate)/3); // or just 36.
	localparam register_size = 36;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample_pos;
	input sample_neg;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 36-bit chirp, split into positive and negative sections
	//localparam PN_Code_pos = 36'b 100000000000000100000000001000000100;
	//localparam PN_Code_neg = 36'b 000000001000000000000100000000100001;	
	
	// Define internal connections 
	// For the hash network
	reg [4:0] hash_register [0:register_size-1]; 
	wire [4:0] new_hashes [0:register_size-1];
	wire hash_adc_sample_pos [0:register_size-2];	
	wire hash_adc_sample_neg [0:register_size-2]; 
	reg hash_update;
	// For the hash count decoder
	wire [2:0] counted_bits [0:7]; 
	wire [outputbits-1:0] corr;
	
	// Connect the first hash generator to the new samples. This one does not need a last sample decoder
	hash_gen hashgen_one (
		.old_hash(hash_register[0]),
		.adc_pos(sample_pos),
		.adc_neg(sample_neg),
		.new_hash(new_hashes[0])
		);
	
	// Create the 5-bit hash network
	genvar k;
	generate
		for(k=1; k<register_size; k=k+1)
			begin : hash_network
				// Connect the first hash generator to the new samples.
				last_sample_dec sample_decoder (
				.hash(hash_register[k-1]),
				.adc_pos(hash_adc_sample_pos[k-1]),
				.adc_neg(hash_adc_sample_neg[k-1])
				);
				
				hash_gen hashgenerator (
				.old_hash(hash_register[k]),
				.adc_pos(hash_adc_sample_pos[k-1]),
				.adc_neg(hash_adc_sample_neg[k-1]),
				.new_hash(new_hashes[k])
				);
			end
	endgenerate	
	
	// Connect the count decoders to the appropriate sections of the hash network, i.e. anywhere with a 1. Need 8 in all 
	// Section of the PN code with +1 use a positive counter, and sections with -1 use a negative counter
	count_decoder_alt counter1 (
		.hash(hash_register[0]),
		.polarity(1'b1),
		.count(counted_bits[0])
		);
	count_decoder_alt counter2 (
		.hash(hash_register[2]),
		.polarity(1'b0),
		.count(counted_bits[1])
		);
	count_decoder_alt counter3 (
		.hash(hash_register[5]),
		.polarity(1'b1),
		.count(counted_bits[2])
		);
	count_decoder_alt counter4 (
		.hash(hash_register[9]),
		.polarity(1'b0),
		.count(counted_bits[3])
		);
	count_decoder_alt counter5 (
		.hash(hash_register[14]),
		.polarity(1'b1),
		.count(counted_bits[4])
		);
	count_decoder_alt counter6 (
		.hash(hash_register[20]),  
		.polarity(1'b0),
		.count(counted_bits[5])
		);
	count_decoder_alt counter7 (
		.hash(hash_register[27]),
		.polarity(1'b1),
		.count(counted_bits[6])
		);
	count_decoder_alt counter8 (
		.hash(hash_register[35]),
		.polarity(1'b0),
		.count(counted_bits[7])
		);
	
	// Sum the output of the hash decoders for the final correlation result
	assign corr = counted_bits[0]+counted_bits[1]+counted_bits[2]+counted_bits[3]+counted_bits[4]+counted_bits[5]+counted_bits[6]+counted_bits[7]; 
	
	// ADC action, clock a new hash into the hash network, based on the new sample 
	integer i;
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin 
					hash_update <= 0;
					// Reset the hash register contents
					begin
						for (i=0; i<register_size; i=i+1) hash_register[i] <= 5'b00000;
					end
					
					// TODO! Find a better way to reset the registers
					//hash_register[0] <= 5'b00000;
					//					hash_register[1] <= 5'b00000;
					//					hash_register[2] <= 5'b00000;
					//					hash_register[3] <= 5'b00000;
					//					hash_register[4] <= 5'b00000;
					//					hash_register[5] <= 5'b00000;
					//					hash_register[6] <= 5'b00000;
					//					hash_register[7] <= 5'b00000;
					//					hash_register[8] <= 5'b00000;
					//					hash_register[9] <= 5'b00000;
					//					
					//					hash_register[10] <= 5'b00000;
					//					hash_register[11] <= 5'b00000;
					//					hash_register[12] <= 5'b00000;
					//					hash_register[13] <= 5'b00000;
					//					hash_register[14] <= 5'b00000;
					//					hash_register[15] <= 5'b00000;
					//					hash_register[16] <= 5'b00000;
					//					hash_register[17] <= 5'b00000;
					//					hash_register[18] <= 5'b00000;
					//					hash_register[19] <= 5'b00000;
					//					
					//					hash_register[20] <= 5'b00000;
					//					hash_register[21] <= 5'b00000;
					//					hash_register[22] <= 5'b00000;
					//					hash_register[23] <= 5'b00000;
					//					hash_register[24] <= 5'b00000;
					//					hash_register[25] <= 5'b00000;
					//					hash_register[26] <= 5'b00000;
					//					hash_register[27] <= 5'b00000;
					//					hash_register[28] <= 5'b00000;
					//					hash_register[29] <= 5'b00000;
					//					
					//					hash_register[30] <= 5'b00000;
					//					hash_register[31] <= 5'b00000;
					//					hash_register[32] <= 5'b00000;
					//					hash_register[33] <= 5'b00000;
					//					hash_register[34] <= 5'b00000;
					//					hash_register[35] <= 5'b00000;
			end else begin
					// Load the new hashes into the hash storage registers
					if (adc_ready) begin 
							// Update the hash values on the next clock cycle
							hash_update <= 1;
					end else if (hash_update) begin
							hash_update <= 0;								 
							
							// Load the new hash values
							begin
								for (i=0; i<register_size; i=i+1) hash_register[i] <= new_hashes[i];
							end
							
							// TODO! Find a better way to update the register
							//hash_register[0] <= new_hashes[0];
							//							hash_register[1] <= new_hashes[1];
							//							hash_register[2] <= new_hashes[2];
							//							hash_register[3] <= new_hashes[3];
							//							hash_register[4] <= new_hashes[4];
							//							hash_register[5] <= new_hashes[5];
							//							hash_register[6] <= new_hashes[6];
							//							hash_register[7] <= new_hashes[7];
							//							hash_register[8] <= new_hashes[8];
							//							hash_register[9] <= new_hashes[9];
							//							
							//							hash_register[10] <= new_hashes[10];
							//							hash_register[11] <= new_hashes[11];
							//							hash_register[12] <= new_hashes[12];
							//							hash_register[13] <= new_hashes[13];
							//							hash_register[14] <= new_hashes[14];
							//							hash_register[15] <= new_hashes[15];
							//							hash_register[16] <= new_hashes[16];
							//							hash_register[17] <= new_hashes[17];
							//							hash_register[18] <= new_hashes[18];
							//							hash_register[19] <= new_hashes[19];
							//							
							//							hash_register[20] <= new_hashes[20];
							//							hash_register[21] <= new_hashes[21];
							//							hash_register[22] <= new_hashes[22];
							//							hash_register[23] <= new_hashes[23];
							//							hash_register[24] <= new_hashes[24];
							//							hash_register[25] <= new_hashes[25];
							//							hash_register[26] <= new_hashes[26];
							//							hash_register[27] <= new_hashes[27];
							//							hash_register[28] <= new_hashes[28];
							//							hash_register[29] <= new_hashes[29];
							//							
							//							hash_register[30] <= new_hashes[30];
							//							hash_register[31] <= new_hashes[31];
							//							hash_register[32] <= new_hashes[32];
							//							hash_register[33] <= new_hashes[33];
							//							hash_register[34] <= new_hashes[34];
							//							hash_register[35] <= new_hashes[35];
						end
				end
		end
	
endmodule
