//-----------------------------------------------------------------------------
//
// Title       : hash_corr_three_muxed_clock
// Design      : hash_corr
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Correlator for Direct Sequence Spread Spectrum radio.
//				Uses a 36-bit chip as the PN code.
//
//				Uses a 5-bit hash to store sample groups of three, which otherwise require
//				6-bits. A hash generator must run through each 5-bit hash to update them,
//				for every new sample recevied.
//
//				Both the count decoders, and the hash generators are multiplexed.
//				The hash register contents are updated on each clock pulse, using
//				a multiplexer for each register to select either it's current
//				contents, or the hash generator output.
//
//				Requires 36 + 8 clock cycles to compute the result.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module hash_corr_three_muxed_clock (reset, clk, sample_pos, sample_neg, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 3;		// NB! Must be 3
	localparam outputbits = 6;
	
	//localparam register_size = ((36*oversamplingrate)/3); // or just 36.
	localparam register_size = 36;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample_pos;
	input sample_neg;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 36-bit chirp, split into positive and negative sections
	//localparam PN_Code_pos = 36'b 100000000000000100000000001000000100;
	//localparam PN_Code_neg = 36'b 000000001000000000000100000000100001;	
	
	// Define internal connections 
	// For the hash network
	reg [4:0] hash_register [0:register_size-1]; 
	reg [4:0] hash_input [0:register_size-1];														 
	
	reg [4:0] old_hash_section, current_hash_section, counted_section;
	wire [4:0] new_hashes, first_newhash;
	wire hash_adc_sample_pos;	
	wire hash_adc_sample_neg; 
	
	// For the hash count decoder 
	reg chip_polarity;
	wire [2:0] counted_chip; 
	wire [outputbits-1:0] corr;
	
	// State machine stuff
	localparam RESET=0, IDLE=1, START=2; 
	reg [6:0] state, nxtState;
	
	// Connect the first hash generator to the new samples. This one does not need a last sample decoder
	hash_gen hashgen_one (
		.old_hash(hash_register[0]),
		.adc_pos(sample_pos),
		.adc_neg(sample_neg),
		.new_hash(first_newhash)
		);
	
	// Connect the hash generator which is muxed around, along with it's last sample decoder
	last_sample_dec sample_decoder (
		.hash(old_hash_section),
		.adc_pos(hash_adc_sample_pos),
		.adc_neg(hash_adc_sample_neg)
		);
	hash_gen hashgenerator (
		.old_hash(current_hash_section),
		.adc_pos(hash_adc_sample_pos),
		.adc_neg(hash_adc_sample_neg),
		.new_hash(new_hashes)
		);
	
	// Connect the hash decoder used to determine the count
	count_decoder counter1 (
		.hash(counted_section),
		.polarity(chip_polarity),
		.count(counted_chip)
		); 						   
	
	assign corr = hash_register[2];
	
	// ADC action, clock a new hash into the hash network, based on the new sample 
	integer i;
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin
					//corr <= 0;
					state <= RESET;	
					// Reset the hash register contents
					begin
						for (i=0; i<register_size; i=i+1) hash_register[i] <= 5'b00000;
					end	
					
			end else begin
					// Load the new hashes into the hash storage registers
					if (adc_ready) begin 										
							// Reset the state machine
							state <= START;
							//corr <= 0;
					end else begin
							// Change state machine state	
							state <= nxtState;	
							// Count the correlation
							//corr <= corr + counted_chip;	
							// Clock in a new hash value
							begin
								for (i=0; i<register_size; i=i+1) hash_register[i] <= hash_input[i];
							end
						end
				end
		end	
	
	// State machine used to multiplex the hash generator
	always @(state) begin
			// Set default conditions
			nxtState = IDLE;
			counted_section = 5'b10111;	// NB! This particular state gives a count of 0	
			begin
				for (i=0; i<register_size; i=i+1) hash_input[i] = hash_register[i];
			end
			
			case (state)
				RESET : begin
						nxtState = IDLE;
					end
				IDLE : begin
						nxtState = IDLE;
					end
				START : begin
						// Load the new hash values	
						hash_input[0] = first_newhash;
						// Prepare for the next hash
						old_hash_section = hash_register[0];
						current_hash_section = hash_register[1];
						
						nxtState = 3;
					end
				3 : begin
						hash_input[1] = new_hashes;
						old_hash_section = hash_register[1];
						current_hash_section = hash_register[2];
						nxtState = 4;
					end
				4 : begin 					   
						hash_input[2] = new_hashes;
						old_hash_section = hash_register[2];
						current_hash_section = hash_register[3];
						nxtState = 5;
					end
				5 : begin
						hash_input[3] = new_hashes;
						old_hash_section = hash_register[3];
						current_hash_section = hash_register[4];
						nxtState = 6;
					end
				6 : begin
						hash_input[4] = new_hashes;
						old_hash_section = hash_register[4];
						current_hash_section = hash_register[5];
						nxtState = 7;
					end
				7 : begin
						hash_input[5] = new_hashes;
						old_hash_section = hash_register[5];
						current_hash_section = hash_register[6];
						nxtState = 8;
					end
				8 : begin
						hash_input[6] = new_hashes;
						old_hash_section = hash_register[6];
						current_hash_section = hash_register[7];
						nxtState = 9;
					end
				9 : begin
						hash_input[7] = new_hashes;
						old_hash_section = hash_register[7];
						current_hash_section = hash_register[8];
						nxtState = 10;
					end
				10 : begin
						hash_input[8] = new_hashes;
						old_hash_section = hash_register[8];
						current_hash_section = hash_register[9];
						nxtState = 11;
					end
				11 : begin
						hash_input[9] = new_hashes;
						old_hash_section = hash_register[9];
						current_hash_section = hash_register[10];
						nxtState = 12;
					end
				12 : begin
						hash_input[10] = new_hashes;
						old_hash_section = hash_register[10];
						current_hash_section = hash_register[11];
						nxtState = 13;
					end
				13 : begin
						hash_input[11] = new_hashes;
						old_hash_section = hash_register[11];
						current_hash_section = hash_register[12];
						nxtState = 14;
					end
				14 : begin
						hash_input[12] = new_hashes;
						old_hash_section = hash_register[12];
						current_hash_section = hash_register[13];
						nxtState = 15;
					end
				15 : begin
						hash_input[13] = new_hashes;
						old_hash_section = hash_register[13];
						current_hash_section = hash_register[14];
						nxtState = 16;
					end
				16 : begin
						hash_input[14] = new_hashes;
						old_hash_section = hash_register[14];
						current_hash_section = hash_register[15];
						nxtState = 17;
					end
				17 : begin
						hash_input[15] = new_hashes;
						old_hash_section = hash_register[15];
						current_hash_section = hash_register[16];
						nxtState = 18;
					end
				18 : begin
						hash_input[16] = new_hashes;
						old_hash_section = hash_register[16];
						current_hash_section = hash_register[17];
						nxtState = 19;
					end
				19 : begin
						hash_input[17] = new_hashes;
						old_hash_section = hash_register[17];
						current_hash_section = hash_register[18];
						nxtState = 20;
					end
				20 : begin
						hash_input[18] = new_hashes;
						old_hash_section = hash_register[18];
						current_hash_section = hash_register[19];
						nxtState = 21;
					end
				21 : begin
						hash_input[19] = new_hashes;
						old_hash_section = hash_register[19];
						current_hash_section = hash_register[20];
						nxtState = 22;
					end
				22 : begin
						hash_input[20] = new_hashes;
						old_hash_section = hash_register[20];
						current_hash_section = hash_register[21];
						nxtState = 23;
					end
				23 : begin
						hash_input[21] = new_hashes;
						old_hash_section = hash_register[21];
						current_hash_section = hash_register[22];
						nxtState = 24;
					end
				24 : begin
						hash_input[22] = new_hashes;
						old_hash_section = hash_register[22];
						current_hash_section = hash_register[23];
						nxtState = 25;
					end
				25 : begin
						hash_input[23] = new_hashes;
						old_hash_section = hash_register[23];
						current_hash_section = hash_register[24];
						nxtState = 26;
					end
				26 : begin
						hash_input[24] = new_hashes;
						old_hash_section = hash_register[24];
						current_hash_section = hash_register[25];
						nxtState = 27;
					end
				27 : begin
						hash_input[25] = new_hashes;
						old_hash_section = hash_register[25];
						current_hash_section = hash_register[26];
						nxtState = 28;
					end
				28 : begin
						hash_input[26] = new_hashes;
						old_hash_section = hash_register[26];
						current_hash_section = hash_register[27];
						nxtState = 29;
					end
				29 : begin
						hash_input[27] = new_hashes;
						old_hash_section = hash_register[27];
						current_hash_section = hash_register[28];
						nxtState = 30;
					end
				30 : begin
						hash_input[28] = new_hashes;
						old_hash_section = hash_register[28];
						current_hash_section = hash_register[29];
						nxtState = 31;
					end
				31 : begin
						hash_input[29] = new_hashes;
						old_hash_section = hash_register[29];
						current_hash_section = hash_register[30];
						nxtState = 32;
					end
				32 : begin
						hash_input[30] = new_hashes;
						old_hash_section = hash_register[30];
						current_hash_section = hash_register[31];
						nxtState = 33;
					end
				33 : begin
						hash_input[31] = new_hashes;
						old_hash_section = hash_register[31];
						current_hash_section = hash_register[32];
						nxtState = 34;
					end
				34 : begin
						hash_input[32] = new_hashes;
						old_hash_section = hash_register[32];
						current_hash_section = hash_register[33];
						nxtState = 35;
					end
				35 : begin
						hash_input[33] = new_hashes;
						old_hash_section = hash_register[33];
						current_hash_section = hash_register[34];
						nxtState = 36;
					end
				36 : begin
						hash_input[34] = new_hashes;
						old_hash_section = hash_register[34];
						current_hash_section = hash_register[35];
						nxtState = 37;
					end
				37 : begin
						hash_input[35] = new_hashes;
						old_hash_section = hash_register[0];
						current_hash_section = hash_register[1];
						nxtState = 64;
					end
				
				// Start muxing the decoder, and sum the correlation result
				64 : begin
						counted_section = hash_register[0];
						chip_polarity = 1'b1; 
						nxtState = 65;
					end
				65 : begin
						counted_section = hash_register[2];
						chip_polarity = 1'b0; 
						nxtState = 66;
					end
				66 : begin
						counted_section = hash_register[5];
						chip_polarity = 1'b1; 
						nxtState = 67;
					end
				67 : begin
						counted_section = hash_register[9];
						chip_polarity = 1'b0; 
						nxtState = 68;
					end
				68 : begin
						counted_section = hash_register[14];
						chip_polarity = 1'b1; 
						nxtState = 69;
					end
				69 : begin
						counted_section = hash_register[20];
						chip_polarity = 1'b0; 
						nxtState = 70;
					end
				70 : begin
						counted_section = hash_register[27];
						chip_polarity = 1'b1; 
						nxtState = 71;
					end
				71 : begin
						counted_section = hash_register[35];
						chip_polarity = 1'b0; 
						nxtState = IDLE;
					end
				
				default : begin
						old_hash_section = hash_register[0];
						counted_section = 5'b10111;
						chip_polarity = 1'b1;
						nxtState = IDLE;
					end
			endcase 
		end
	
endmodule
