//-----------------------------------------------------------------------------
//
// Title       : twobit_corr_mux_eight
// Design      : parallel_correlator
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses a 36-bit chip as the PN code. ADC sampling rate can be set.
//				Output result is ready some time before the next sample is expected.
//				
//				Uses a single 8-bit parallel counter.
//
//				The state machine must be scaled manually to match the sampling rate.
//				The correlation result is ready (1 + N*2) clock cycles after a new
//				ADC sample.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module twobit_corr_mux_eight (reset, clk, sample_pos, sample_neg, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 4;		// NB! Must be 2 or greater
	localparam outputbits = 7;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample_pos;
	input sample_neg;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 36-bit chirp, split into positive and negative sections
	//localparam PN_Code_pos = 36'b 100000000000000100000000001000000100;
	//localparam PN_Code_neg = 36'b 000000001000000000000100000000100001;	
	
	// Define internal connections 
	reg [outputbits-1:0] corr;
	reg [(36*oversamplingrate)-1:0] adc_samples_pos;	
	reg [(36*oversamplingrate)-1:0] adc_samples_neg;  
	wire [(16*oversamplingrate)-1:0] count_bits;
	
	// Sum the output of every sample multiplication to get the correlation measure.
	reg [7:0] counted_section;
	wire [3:0] bitsum; 
	
	// Connect the parallel counter
	eight_bit_cnter bitcounter (
		.in(counted_section),
		.sum(bitsum) 
		);
	
	// State machine stuff
	localparam IDLE=0, START=1, WAIT=2; 
	reg [3:0] state, nxtState;
	
	// Route the correct sections of the ADC registers to a single bus which can be counted
	genvar k;
	generate
		for(k=0; k<oversamplingrate; k=k+1)
			begin : parallel_add
				// First the sections containing the PN code
				assign count_bits[3+k*16:0+k*16] = {adc_samples_pos[2*oversamplingrate+k], adc_samples_pos[9*oversamplingrate+k], adc_samples_pos[20*oversamplingrate+k], adc_samples_pos[35*oversamplingrate+k]};
				assign count_bits[7+k*16:4+k*16] = {adc_samples_neg[k], adc_samples_neg[5*oversamplingrate+k], adc_samples_neg[14*oversamplingrate+k], adc_samples_neg[27*oversamplingrate+k]};
				
				// Same as above, just with the invese of the PN code
				not(count_bits[8+k*16], adc_samples_pos[k]);
				not(count_bits[9+k*16], adc_samples_pos[5*oversamplingrate+k]);
				not(count_bits[10+k*16], adc_samples_pos[14*oversamplingrate+k]);
				not(count_bits[11+k*16], adc_samples_pos[27*oversamplingrate+k]);
				
				not(count_bits[12+k*16], adc_samples_neg[2*oversamplingrate+k]);
				not(count_bits[13+k*16], adc_samples_neg[9*oversamplingrate+k]);
				not(count_bits[14+k*16], adc_samples_neg[20*oversamplingrate+k]);
				not(count_bits[15+k*16], adc_samples_neg[35*oversamplingrate+k]);														  
				
			end
	endgenerate	
	
	// ADC action, clock a new value into the shift registers 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin
					corr <= 0;
					state <= IDLE;
					adc_samples_pos <= 0;  
					adc_samples_neg <= 0;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples_pos <= {adc_samples_pos[(36*oversamplingrate)-2:0], sample_pos};	
							adc_samples_neg <= {adc_samples_neg[(36*oversamplingrate)-2:0], sample_neg};
							// Reset the state machine
							corr <= 0; 
							state <= WAIT;
					end else begin
							// Change state machine state	
							state <= nxtState;
							corr <= corr + bitsum; 
						end
				end
		end	
	
	// State machine used to multiplex the four bit parallel counter
	always @(state) begin	  
			nxtState = IDLE;
			case (state)
				IDLE : begin
						counted_section = 8'b0000_0000;
						nxtState = IDLE;
					end
				WAIT : begin
						// Wait a clock pulse so the ADC register can shift
						counted_section = 8'b0000_0000;
						nxtState = START;
					end
				START : begin
						counted_section = count_bits[7:0];
						nxtState = 3;
					end
				3 : begin
						counted_section = count_bits[15:8];
						nxtState = 4;
					end
				4 : begin
						counted_section = count_bits[23:16];
						nxtState = 5;
					end 
				5 : begin
						counted_section = count_bits[31:24];
						nxtState = 6;
					end
				6 : begin
						counted_section = count_bits[39:32];
						nxtState = 7;
					end
				7 : begin
						counted_section = count_bits[47:40];
						nxtState = 8;
					end
				8 : begin
						counted_section = count_bits[55:48];
						nxtState = 9;
					end
				9 : begin
						counted_section = count_bits[63:56];
						nxtState = IDLE;
					end
				
				default : begin
						counted_section = 8'b0000_0000;
						nxtState = IDLE;
					end
			endcase 
		end
	
endmodule
