//-----------------------------------------------------------------------------
//
// Title       : twobit_corr_ref_sixteen
// Design      : parallel_correlator
// Author      : Eirik Taylor, 2014
// Company     : 
//
//
//-----------------------------------------------------------------------------
//
// Description : Parallel correlator (matched filter) for Direct Sequence Spread Spectrum radio.
//				Uses a 36-bit chip as the PN code. ADC sampling rate can be set.
//				Output result is ready some time before the next sample is expected.
//				Setting the sampling rate automatically scales the design
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

module twobit_corr_ref_sixteen (reset, clk, sample_pos, sample_neg, corr, adc_ready);
	// Define some correlator properties
	localparam oversamplingrate = 5;		// NB! Must be 2 or greater
	localparam outputbits = 7;
	
	// Define I/O directions
	output [outputbits-1:0] corr;
	input reset;	
	input clk;	
	input sample_pos;
	input sample_neg;
	input adc_ready;
	
	// Define the PN code used to correlate against
	// 36-bit chirp, split into positive and negative sections
	//localparam PN_Code_pos = 36'b 100000000000000100000000001000000100;
	//localparam PN_Code_neg = 36'b 000000001000000000000100000000100001;	
	
	// Define internal connections 
	wire [outputbits-1:0] corr;
	reg [(36*oversamplingrate)-1:0] adc_samples_pos;	
	reg [(36*oversamplingrate)-1:0] adc_samples_neg;
	
	// Sum the output of every sample multiplication to get the correlation measure.  
	wire [4:0] bitsum [0:(oversamplingrate-1)];
	
	// Used for the inverse of the PN Code
	wire [3:0] positive_zeros [0:oversamplingrate-1]; 
	wire [3:0] negative_zeros [0:oversamplingrate-1];
	
	// For final summation	
	wire [outputbits-1:0] partial_sum [0:oversamplingrate-1];
	
	// Connect the required number of 16-bit parallel counters to the adc registers
	genvar k;
	generate
		for(k=0; k<oversamplingrate; k=k+1)
			begin : parallel_add	
				// Connect the parallel counter to the positive samples, where there must be zeros
				not(positive_zeros[k][0], adc_samples_pos[k]);
				not(positive_zeros[k][1], adc_samples_pos[5*oversamplingrate+k]);
				not(positive_zeros[k][2], adc_samples_pos[14*oversamplingrate+k]);
				not(positive_zeros[k][3], adc_samples_pos[27*oversamplingrate+k]);   
				// Connect the parallel counter to the negative samples, where there must be zeros
				not(negative_zeros[k][0], adc_samples_neg[2*oversamplingrate+k]);
				not(negative_zeros[k][1], adc_samples_neg[9*oversamplingrate+k]);
				not(negative_zeros[k][2], adc_samples_neg[20*oversamplingrate+k]);
				not(negative_zeros[k][3], adc_samples_neg[35*oversamplingrate+k]);	
				// Connect the parallel counter to the positive samples
				sixteen_bit_cnter bitcounter_pos (
				.in({adc_samples_pos[2*oversamplingrate+k], adc_samples_pos[9*oversamplingrate+k], adc_samples_pos[20*oversamplingrate+k], adc_samples_pos[35*oversamplingrate+k], positive_zeros[k], adc_samples_neg[k], adc_samples_neg[5*oversamplingrate+k], adc_samples_neg[14*oversamplingrate+k], adc_samples_neg[27*oversamplingrate+k], negative_zeros[k]}),
				.sum(bitsum[k][4:0]) 
				); 
			end
	endgenerate	
	
	// Automatically sum each partial product. Use a tree-like structure
	assign partial_sum[0] = bitsum[0];
	genvar j;
	generate 
		for(j=1; j<oversamplingrate; j=j+1)
			begin : partial_add
				assign partial_sum[j] = bitsum[j] + partial_sum[j-1];  
			end
	endgenerate
	// The final sum is equal to the last indexed partial sum
	assign corr = partial_sum[oversamplingrate-1];
	
	// ADC action, clock a new value into the shift registers 
	always @ (posedge clk or negedge reset)
		begin
			if (!reset) begin 
					adc_samples_pos <= 0;  
					adc_samples_neg <= 0;
			end else begin
					// Shift the ADC reg
					if (adc_ready) begin 
							adc_samples_pos <= {adc_samples_pos[(36*oversamplingrate)-2:0], sample_pos};	
							adc_samples_neg <= {adc_samples_neg[(36*oversamplingrate)-2:0], sample_neg};
						end
				end
		end
	
endmodule
